/*
Owner: STDIO Training
Author: Phan Tan Phuc
Website: https://training.stdio.vn
Email: developer@stdio.vn
*/

#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <string.h>

typedef unsigned char byte;

int _width;
int _height;
int _bitdepth;

typedef struct
{
	char			m_type[2];
	unsigned char	m_fSize[4];
	unsigned char	m_reserved1[2];
	unsigned char	m_reserved2[2];
	unsigned char	m_data_offset[4];
}BitMapHeader;

typedef struct
{
	unsigned char m_size[4];
	unsigned char m_width[4];
	unsigned char m_height[4];
	unsigned char m_planes[2];
	unsigned char m_bitDepth[2];
	unsigned char m_compression_type[16];
	unsigned char m_color_used[4];
	unsigned char m_color_important[4];
}BitMapInformation;

void readImage(const char * _filePath, BitMapHeader & _header, BitMapInformation & _information, unsigned char * &_data)
{
	FILE* _imageFile;
	_imageFile = fopen(_filePath, "rb");

	if (_imageFile == nullptr)
	{
		return;
	}


	fread(&_header, sizeof(BitMapHeader), sizeof(char), _imageFile);
	fread(&_information, sizeof(BitMapInformation), sizeof(char), _imageFile);

	_width = *(int*)_information.m_width;
	_height = *(int*)_information.m_height;
	_bitdepth = *(int*)_information.m_bitDepth;
	int data_offset = *(int*)_header.m_data_offset;

	int _rowSize = (_bitdepth * _width + 31) / 32 * 4;

	if (_data != nullptr)
	{
		delete[] _data;
		_data = nullptr;
	}
	_data = new unsigned char[_rowSize * _height];


	fseek(_imageFile, data_offset, SEEK_SET);
	fread(_data, _rowSize * _height, sizeof(char), _imageFile);

	fclose(_imageFile);
}

void corr(byte* data, int width, int height, int byte_per_pixel, float* kernel, int kernel_size) {

	int padding = kernel_size / 2;
	int width_size = width * byte_per_pixel;
	int height_size = height * byte_per_pixel;

	int _rowSize = (byte_per_pixel * 8 * width + 31) / 32 * 4;

	int img_size = width * height * byte_per_pixel * sizeof(byte);

	byte* temp_img = new byte[_rowSize * height];

	for (int r = padding; r < height - padding; ++r) {
		for (int c = padding; c < width - padding; ++c) {

			float sum_c1 = 0.0f;
			float sum_c2 = 0.0f;
			float sum_c3 = 0.0f;

			for (int kr = -padding; kr <= padding; ++kr) {
				for (int kc = -padding; kc <= padding; ++kc) {
					int r_corr = r + kr;
					int c_corr = c + kc;
					sum_c1 += data[_rowSize * r_corr + c_corr * byte_per_pixel + 0] * kernel[kernel_size * (kr + padding) + (kc + padding)];
					sum_c2 += data[_rowSize * r_corr + c_corr * byte_per_pixel + 1] * kernel[kernel_size * (kr + padding) + (kc + padding)];
					sum_c3 += data[_rowSize * r_corr + c_corr * byte_per_pixel + 2] * kernel[kernel_size * (kr + padding) + (kc + padding)];
				}
			}

			if (sum_c1 > 255) sum_c1 = 255;
			if (sum_c1 < 0) sum_c1 = 0;
			if (sum_c2 > 255) sum_c2 = 255;
			if (sum_c2 < 0) sum_c2 = 0;
			if (sum_c3 > 255) sum_c3 = 255;
			if (sum_c3 < 0) sum_c3 = 0;

			temp_img[_rowSize * r + c * byte_per_pixel + 0] = (byte)sum_c1;
			temp_img[_rowSize * r + c * byte_per_pixel + 1] = (byte)sum_c2;
			temp_img[_rowSize * r + c * byte_per_pixel + 2] = (byte)sum_c3;
		}
	}

	memcpy(data, temp_img, img_size);
	delete[] temp_img;


}

void saveImage(const char * _filePath, BitMapHeader & _header, BitMapInformation & _information, const unsigned char * _data)
{
	int _width = *(int*)_information.m_width;
	int _height = *(int*)_information.m_height;
	int _bitdepth = *(int*)_information.m_bitDepth;
	int _offset = *(int*)_header.m_data_offset;

	FILE * _imageFile;
	_imageFile = fopen(_filePath, "wb");

	fwrite(&_header, sizeof(BitMapHeader), sizeof(char), _imageFile);
	fwrite(&_information, sizeof(BitMapInformation), 1, _imageFile);

	int _size = (_bitdepth * _width + 31) / 32 * 4;

	fseek(_imageFile, _offset, SEEK_SET);
	fwrite(_data, _size * _height, sizeof(char), _imageFile);
	fclose(_imageFile);
}

int main()
{
	// initilize
	byte * _dataImage = nullptr;

	BitMapHeader		_header;
	BitMapInformation	_infomation;

	// Thay STDIO_Image bang duong dan den anh Bitmap phu hop

	readImage("STDIO_Image.bmp", _header, _infomation, _dataImage);

	float laplacekernel[] =
		{ -1, -1, -1,
		  -1, 8, -1,
		  -1, -1, -1 };

	corr(_dataImage, _width, _height, _bitdepth / 8, laplacekernel, 3);

	saveImage("STDIO_ImageNew.bmp", _header, _infomation, _dataImage);

	delete[] _dataImage;

	return 0;
}